// QR CODE GENERATOR
//
// Get the latest version here:
//      https://viereck.ch/qr-code
//
// The code below was adapted from https://kazuhikoarase.github.io (Kazuhiko Arase).
// It is functionally equivalent, but the code was greatly restructured.
//
// Author: Thomas Lochmatter, thomas.lochmatter@viereck.ch
// License: MIT

// *** Conversion between strings and Uint8Arrays (using UTF-8 encoding) ***

// Returns the number of bytes required to encode a UTF-8 char.
function utf8CharLength(charCode) {	// Uint ⇒ Uint
	return charCode < 0x80 ? 1 : charCode < 0x800 ? 2 : charCode < 0x10000 ? 3 : charCode < 0x200000 ? 4 : charCode < 0x4000000 ? 5 : 6;
}

// Adds a single char to a byte array at position "offset". Returns the new offset after adding the char.
function addUtf8CharToBytes(bytes, offset, charCode) {	// Uint8Array, Uint, Uint ⇒ Uint
	var position = offset;
	if (charCode < 0x80) {
		bytes[position++] = charCode;
	} else if (charCode < 0x800) {
		bytes[position++] = 0xc0 | (charCode >>> 6);
		bytes[position++] = 0x80 | (charCode & 0x3f);
	} else if (charCode < 0x10000) {
		bytes[position++] = 0xe0 | (charCode >>> 12);
		bytes[position++] = 0x80 | ((charCode >>> 6) & 0x3f);
		bytes[position++] = 0x80 | (charCode & 0x3f);
	} else if (charCode < 0x200000) {
		bytes[position++] = 0xf0 | (charCode >>> 18);
		bytes[position++] = 0x80 | ((charCode >>> 12) & 0x3f);
		bytes[position++] = 0x80 | ((charCode >>> 6) & 0x3f);
		bytes[position++] = 0x80 | (charCode & 0x3f);
	} else if (charCode < 0x4000000) {
		// Must never appear in a valid UTF-8 sequence
		bytes[position++] = 0xf8 | (charCode >>> 24);
		bytes[position++] = 0x80 | ((charCode >>> 18) & 0x3f);
		bytes[position++] = 0x80 | ((charCode >>> 12) & 0x3f);
		bytes[position++] = 0x80 | ((charCode >>> 6) & 0x3f);
		bytes[position++] = 0x80 | (charCode & 0x3f);
	} else {
		// Must never appear in a valid UTF-8 sequence
		bytes[position++] = 0xfc | (charCode >>> 30);
		bytes[position++] = 0x80 | ((charCode >>> 24) & 0x3f);
		bytes[position++] = 0x80 | ((charCode >>> 18) & 0x3f);
		bytes[position++] = 0x80 | ((charCode >>> 12) & 0x3f);
		bytes[position++] = 0x80 | ((charCode >>> 6) & 0x3f);
		bytes[position++] = 0x80 | (charCode & 0x3f);
	}
	return position;
}

// Converts a string into a byte array using UTF-8 encoding.
function bytesFromText(text) {	// String ⇒ Uint8Array
	var length = 0;
	for (var i = 0; i < text.length; i++)
		length += utf8CharLength(text.charCodeAt(i));

	var offset = 0;
	var bytes = new Uint8Array(length);
	for (var i = 0; i < text.length; i++)
		offset = addUtf8CharToBytes(bytes, offset, text.charCodeAt(i));

	return bytes;
};

// Converts a byte array to a string, interpreting the bytes as UTF-8 sequence. Invalid bytes are silently discarded.
function textFromBytes(bytes) {	// Uint8Array ⇒ String
	var text = '';
	var offset = 0;
	while (offset < bytes.length) {
		var byte = bytes[offset];
		if (byte < 0x80) {
			text += String.fromCharCode(byte);
			offset += 1;
		} else if (byte < 0xc0) {
			// continuation byte, ignore
			offset += 1;
		} else if (byte < 0xe0) {
			if (offset + 2 > bytes.length) break;
			var charCode = (byte & 0x1f) << 6;
			charCode |= bytes[offset + 1] & 0x7f;
			text += String.fromCharCode(charCode);
			offset += 2;
		} else if (byte < 0xf0) {
			if (offset + 3 > bytes.length) break;
			var charCode = (byte & 0xf) << 12;
			charCode |= (bytes[offset + 1] & 0x7f) << 6;
			charCode |= bytes[offset + 2] & 0x7f;
			text += String.fromCharCode(charCode);
			offset += 3;
		} else if (byte < 0xf8) {
			if (offset + 4 > bytes.length) break;
			var charCode = (byte & 0x7) << 18;
			charCode |= (bytes[offset + 1] & 0x7f) << 12;
			charCode |= (bytes[offset + 2] & 0x7f) << 6;
			charCode |= bytes[offset + 3] & 0x7f;
			text += String.fromCharCode(charCode);
			offset += 4;
		} else if (byte < 0xfc) {
			// Must never appear in a valid UTF-8 sequence
			if (offset + 5 > bytes.length) break;
			var charCode = (byte & 0x3) << 24;
			charCode |= (bytes[offset + 1] & 0x7f) << 18;
			charCode |= (bytes[offset + 2] & 0x7f) << 12;
			charCode |= (bytes[offset + 3] & 0x7f) << 6;
			charCode |= bytes[offset + 4] & 0x7f;
			text += String.fromCharCode(charCode);
			offset += 5;
		} else if (byte < 0xfe) {
			// Must never appear in a valid UTF-8 sequence
			if (offset + 6 > bytes.length) break;
			var charCode = (byte & 0x1) << 30;
			charCode |= (bytes[offset + 1] & 0x7f) << 24;
			charCode |= (bytes[offset + 2] & 0x7f) << 18;
			charCode |= (bytes[offset + 3] & 0x7f) << 12;
			charCode |= (bytes[offset + 4] & 0x7f) << 6;
			charCode |= bytes[offset + 5] & 0x7f;
			text += String.fromCharCode(charCode);
			offset += 6;
		} else {
			// Must never appear in a valid UTF-8 sequence
			offset += 1;
		}
	}

	return text;
};
// QR CODE GENERATOR
//
// Get the latest version here:
//      https://viereck.ch/qr-code
//
// The code below was adapted from https://kazuhikoarase.github.io (Kazuhiko Arase).
// It is functionally equivalent, but the code was greatly restructured.
//
// Author: Thomas Lochmatter, thomas.lochmatter@viereck.ch
// License: MIT

function QRCodeGenerator() {
	var PAD0 = 0xEC;
	var PAD1 = 0x11;

	var NUMBER = this.NUMBER = new Mode(1 << 0, [10, 12, 14]);
	var ALPHANUMERIC = this.ALPHANUMERIC = new Mode(1 << 1, [9, 11, 13]);
	var BYTES = this.BYTES = new Mode(1 << 2, [8, 16, 16]);
	var KANJI = this.KANJI = new Mode(1 << 3, [8, 10, 12]);

	function Mode(code, bitLengths) {
		this.code = code;

		this.getBitSizeOfLengthField = function(typeNumber) {
			if (typeNumber < 10) return bitLengths[0];
			if (typeNumber < 27) return bitLengths[1];
			return bitLengths[2];
		};
	}

	var LOW_REDUNDANCY = this.LOW_REDUNDANCY = new ErrorCorrectionLevel(1, [
		  [b(1, 26, 19)],
		  [b(1, 44, 34)],
		  [b(1, 70, 55)],
		  [b(1, 100, 80)],
		  [b(1, 134, 108)],
		  [b(2, 86, 68)],
		  [b(2, 98, 78)],
		  [b(2, 121, 97)],
		  [b(2, 146, 116)],
		  [b(2, 86, 68), b(2, 87, 69)],
		  [b(4, 101, 81)],
		  [b(2, 116, 92), b(2, 117, 93)],
		  [b(4, 133, 107)],
		  [b(3, 145, 115), b(1, 146, 116)],
		  [b(5, 109, 87), b(1, 110, 88)],
		  [b(5, 122, 98), b(1, 123, 99)],
		  [b(1, 135, 107), b(5, 136, 108)],
		  [b(5, 150, 120), b(1, 151, 121)],
		  [b(3, 141, 113), b(4, 142, 114)],
		  [b(3, 135, 107), b(5, 136, 108)],
		  [b(4, 144, 116), b(4, 145, 117)],
		  [b(2, 139, 111), b(7, 140, 112)],
		  [b(4, 151, 121), b(5, 152, 122)],
		  [b(6, 147, 117), b(4, 148, 118)],
		  [b(8, 132, 106), b(4, 133, 107)],
		  [b(10, 142, 114), b(2, 143, 115)],
		  [b(8, 152, 122), b(4, 153, 123)],
		  [b(3, 147, 117), b(10, 148, 118)],
		  [b(7, 146, 116), b(7, 147, 117)],
		  [b(5, 145, 115), b(10, 146, 116)],
		  [b(13, 145, 115), b(3, 146, 116)],
		  [b(17, 145, 115)],
		  [b(17, 145, 115), b(1, 146, 116)],
		  [b(13, 145, 115), b(6, 146, 116)],
		  [b(12, 151, 121), b(7, 152, 122)],
		  [b(6, 151, 121), b(14, 152, 122)],
		  [b(17, 152, 122), b(4, 153, 123)],
		  [b(4, 152, 122), b(18, 153, 123)],
		  [b(20, 147, 117), b(4, 148, 118)],
		  [b(19, 148, 118), b(6, 149, 119)]
		]);

	var MEDIUM_REDUNDANCY = this.MEDIUM_REDUNDANCY = new ErrorCorrectionLevel(0, [
		  [b(1, 26, 16)],
		  [b(1, 44, 28)],
		  [b(1, 70, 44)],
		  [b(2, 50, 32)],
		  [b(2, 67, 43)],
		  [b(4, 43, 27)],
		  [b(4, 49, 31)],
		  [b(2, 60, 38), b(2, 61, 39)],
		  [b(3, 58, 36), b(2, 59, 37)],
		  [b(4, 69, 43), b(1, 70, 44)],
		  [b(1, 80, 50), b(4, 81, 51)],
		  [b(6, 58, 36), b(2, 59, 37)],
		  [b(8, 59, 37), b(1, 60, 38)],
		  [b(4, 64, 40), b(5, 65, 41)],
		  [b(5, 65, 41), b(5, 66, 42)],
		  [b(7, 73, 45), b(3, 74, 46)],
		  [b(10, 74, 46), b(1, 75, 47)],
		  [b(9, 69, 43), b(4, 70, 44)],
		  [b(3, 70, 44), b(11, 71, 45)],
		  [b(3, 67, 41), b(13, 68, 42)],
		  [b(17, 68, 42)],
		  [b(17, 74, 46)],
		  [b(4, 75, 47), b(14, 76, 48)],
		  [b(6, 73, 45), b(14, 74, 46)],
		  [b(8, 75, 47), b(13, 76, 48)],
		  [b(19, 74, 46), b(4, 75, 47)],
		  [b(22, 73, 45), b(3, 74, 46)],
		  [b(3, 73, 45), b(23, 74, 46)],
		  [b(21, 73, 45), b(7, 74, 46)],
		  [b(19, 75, 47), b(10, 76, 48)],
		  [b(2, 74, 46), b(29, 75, 47)],
		  [b(10, 74, 46), b(23, 75, 47)],
		  [b(14, 74, 46), b(21, 75, 47)],
		  [b(14, 74, 46), b(23, 75, 47)],
		  [b(12, 75, 47), b(26, 76, 48)],
		  [b(6, 75, 47), b(34, 76, 48)],
		  [b(29, 74, 46), b(14, 75, 47)],
		  [b(13, 74, 46), b(32, 75, 47)],
		  [b(40, 75, 47), b(7, 76, 48)],
		  [b(18, 75, 47), b(31, 76, 48)]
		]);

	var QUARTILE_REDUNDANCY = this.QUARTILE_REDUNDANCY = new ErrorCorrectionLevel(3, [
		  [b(1, 26, 13)],
		  [b(1, 44, 22)],
		  [b(2, 35, 17)],
		  [b(2, 50, 24)],
		  [b(2, 33, 15), b(2, 34, 16)],
		  [b(4, 43, 19)],
		  [b(2, 32, 14), b(4, 33, 15)],
		  [b(4, 40, 18), b(2, 41, 19)],
		  [b(4, 36, 16), b(4, 37, 17)],
		  [b(6, 43, 19), b(2, 44, 20)],
		  [b(4, 50, 22), b(4, 51, 23)],
		  [b(4, 46, 20), b(6, 47, 21)],
		  [b(8, 44, 20), b(4, 45, 21)],
		  [b(11, 36, 16), b(5, 37, 17)],
		  [b(5, 54, 24), b(7, 55, 25)],
		  [b(15, 43, 19), b(2, 44, 20)],
		  [b(1, 50, 22), b(15, 51, 23)],
		  [b(17, 50, 22), b(1, 51, 23)],
		  [b(17, 47, 21), b(4, 48, 22)],
		  [b(15, 54, 24), b(5, 55, 25)],
		  [b(17, 50, 22), b(6, 51, 23)],
		  [b(7, 54, 24), b(16, 55, 25)],
		  [b(11, 54, 24), b(14, 55, 25)],
		  [b(11, 54, 24), b(16, 55, 25)],
		  [b(7, 54, 24), b(22, 55, 25)],
		  [b(28, 50, 22), b(6, 51, 23)],
		  [b(8, 53, 23), b(26, 54, 24)],
		  [b(4, 54, 24), b(31, 55, 25)],
		  [b(1, 53, 23), b(37, 54, 24)],
		  [b(15, 54, 24), b(25, 55, 25)],
		  [b(42, 54, 24), b(1, 55, 25)],
		  [b(10, 54, 24), b(35, 55, 25)],
		  [b(29, 54, 24), b(19, 55, 25)],
		  [b(44, 54, 24), b(7, 55, 25)],
		  [b(39, 54, 24), b(14, 55, 25)],
		  [b(46, 54, 24), b(10, 55, 25)],
		  [b(49, 54, 24), b(10, 55, 25)],
		  [b(48, 54, 24), b(14, 55, 25)],
		  [b(43, 54, 24), b(22, 55, 25)],
		  [b(34, 54, 24), b(34, 55, 25)]
		]);

	var HIGH_REDUNDANCY = this.HIGH_REDUNDANCY = new ErrorCorrectionLevel(2, [
		  [b(1, 26, 9)],
		  [b(1, 44, 16)],
		  [b(2, 35, 13)],
		  [b(4, 25, 9)],
		  [b(2, 33, 11), b(2, 34, 12)],
		  [b(4, 43, 15)],
		  [b(4, 39, 13), b(1, 40, 14)],
		  [b(4, 40, 14), b(2, 41, 15)],
		  [b(4, 36, 12), b(4, 37, 13)],
		  [b(6, 43, 15), b(2, 44, 16)],
		  [b(3, 36, 12), b(8, 37, 13)],
		  [b(7, 42, 14), b(4, 43, 15)],
		  [b(12, 33, 11), b(4, 34, 12)],
		  [b(11, 36, 12), b(5, 37, 13)],
		  [b(11, 36, 12)],
		  [b(3, 45, 15), b(13, 46, 16)],
		  [b(2, 42, 14), b(17, 43, 15)],
		  [b(2, 42, 14), b(19, 43, 15)],
		  [b(9, 39, 13), b(16, 40, 14)],
		  [b(15, 43, 15), b(10, 44, 16)],
		  [b(19, 46, 16), b(6, 47, 17)],
		  [b(34, 37, 13)],
		  [b(16, 45, 15), b(14, 46, 16)],
		  [b(30, 46, 16), b(2, 47, 17)],
		  [b(22, 45, 15), b(13, 46, 16)],
		  [b(33, 46, 16), b(4, 47, 17)],
		  [b(12, 45, 15), b(28, 46, 16)],
		  [b(11, 45, 15), b(31, 46, 16)],
		  [b(19, 45, 15), b(26, 46, 16)],
		  [b(23, 45, 15), b(25, 46, 16)],
		  [b(23, 45, 15), b(28, 46, 16)],
		  [b(19, 45, 15), b(35, 46, 16)],
		  [b(11, 45, 15), b(46, 46, 16)],
		  [b(59, 46, 16), b(1, 47, 17)],
		  [b(22, 45, 15), b(41, 46, 16)],
		  [b(2, 45, 15), b(64, 46, 16)],
		  [b(24, 45, 15), b(46, 46, 16)],
		  [b(42, 45, 15), b(32, 46, 16)],
		  [b(10, 45, 15), b(67, 46, 16)],
		  [b(20, 45, 15), b(61, 46, 16)]
		]);

	function b(count, totalCount, dataCount) {
		return {count, totalCount, dataCount};
	}

	function ErrorCorrectionLevel(code, rsBlockTable) {
		this.code = code;

		function estimateTypeNumber(length) {
			for (var i = 0; i < rsBlockTable.length; i++) {
				var typeNumber = i + 1;
				var maxLength = getMaximumLength(typeNumber);
				if (length <= maxLength) return typeNumber;
			}

			return i + 1;
		}

		function getMaximumLength(typeNumber) {
			var maximumLength = 0;
			var rsBlocks = rsBlockTable[typeNumber - 1];
			for (var i = 0; i < rsBlocks.length; i++) {
				var rsBlock = rsBlocks[i];
				maximumLength += rsBlock.count * rsBlock.dataCount;
			}

			return maximumLength;
		}

		function getRSBlocks(typeNumber) {
			var list = [];
			var rsBlocks = rsBlockTable[typeNumber - 1];
			for (var i = 0; i < rsBlocks.length; i++) {
				var rsBlock = rsBlocks[i];
				for (var j = 0; j < rsBlock.count; j++)
					list.push(rsBlock);
			}

			return list;
		}

		this.createData = function(dataList) {
			var minimalLength = 0;
			for (var i = 0; i < dataList.length; i++) {
				var data = dataList[i];
				minimalLength += data.length;
			}

			var typeNumber = estimateTypeNumber(minimalLength);
			while (true) {
				if (typeNumber >= rsBlockTable.length) return null; // Data too long.

				var buffer = createBuffer(typeNumber);
				var maximumBits = getMaximumLength(typeNumber) * 8;
				if (buffer.getLengthInBits() <= maximumBits) break;

				typeNumber += 1;
			}

			function createBuffer(typeNumber) {
				var buffer = new BitBuffer();
				for (var i = 0; i < dataList.length; i++) {
					var data = dataList[i];
					buffer.put(data.mode.code, 4);
					buffer.put(data.length, data.mode.getBitSizeOfLengthField(typeNumber));
					data.writeTo(buffer);
				}

				return buffer;
			}

			if (buffer.getLengthInBits() + 4 <= maximumBits)
				buffer.put(0, 4);

			while (buffer.getLengthInBits() % 8 != 0)
				buffer.putBit(false);

			while (true) {
				if (buffer.getLengthInBits() >= maximumBits) break;
				buffer.put(PAD0, 8);
				if (buffer.getLengthInBits() >= maximumBits) break;
				buffer.put(PAD1, 8);
			}

			var offset = 0;
			var maxDcCount = 0;
			var maxEcCount = 0;
			var rsBlocks = getRSBlocks(typeNumber);
			var dcdata = new Array(rsBlocks.length);
			var ecdata = new Array(rsBlocks.length);
			for (var r = 0; r < rsBlocks.length; r++) {
				var dcCount = rsBlocks[r].dataCount;
				var ecCount = rsBlocks[r].totalCount - dcCount;
				maxDcCount = Math.max(maxDcCount, dcCount);
				maxEcCount = Math.max(maxEcCount, ecCount);
				dcdata[r] = new Array(dcCount);
				for (var i = 0; i < dcdata[r].length; i++)
					dcdata[r][i] = 0xff & buffer.buffer[i + offset];

				offset += dcCount;
				var rsPoly = getErrorCorrectionPolynomial(ecCount);
				var rawPoly = new QRPolynomial(dcdata[r], rsPoly.getLength() - 1);
				var modPoly = rawPoly.mod(rsPoly);
				ecdata[r] = new Array(rsPoly.getLength() - 1);
				for (var i = 0; i < ecdata[r].length; i++) {
					var modIndex = i + modPoly.getLength() - ecdata[r].length;
					ecdata[r][i] = (modIndex >= 0) ? modPoly.get(modIndex) : 0;
				}
			}

			var totalCodeCount = 0;
			for (var i = 0; i < rsBlocks.length; i++)
				totalCodeCount += rsBlocks[i].totalCount;

			var data = new Array(totalCodeCount);
			var index = 0;
			for (var i = 0; i < maxDcCount; i++)
				for (var r = 0; r < rsBlocks.length; r++)
					if (i < dcdata[r].length)
						data[index++] = dcdata[r][i];

			for (var i = 0; i < maxEcCount; i++)
				for (var r = 0; r < rsBlocks.length; r++)
					if (i < ecdata[r].length)
						data[index++] = ecdata[r][i];

			return {data, typeNumber};
		};
	}

	function BitBuffer() {
		this.buffer = [0];
		var lengthInBits = 0;
		var mask = 0x80;

		this.put = function(value, length) {
			var mask = 1 << length;
			for (var i = 0; i < length; i++) {
				mask = mask >>> 1;
				this.putBit((value & mask) == mask);
			}
		};

		this.getLengthInBits = function() {
			return lengthInBits;
		};

		this.putBit = function(bit) {
			if (mask == 0) {
				mask = 0x80;
				this.buffer.push(0);
			}

			if (bit) this.buffer[this.buffer.length - 1] |= mask;

			mask = mask >>> 1;
			lengthInBits += 1;
		};
	}

	function TextData(text) {
		var bytes = bytesFromText(text);

		if (bytes.length != text.length) {
			var extendedBytes = new Uint8Array(bytes.length + 3);
			extendedBytes[0] = 239;
			extendedBytes[1] = 187;
			extendedBytes[2] = 191;
			extendedBytes.set(bytes, 3);
			bytes = extendedBytes;
		}

		this.mode = BYTES;
		this.length = bytes.length;

		this.writeTo = function(buffer) {
			for (var i = 0; i < bytes.length; i++)
				buffer.put(bytes[i], 8);
		};
	}

	var maskFunctions = [
		function(i, j) { return (i + j) % 2 == 0; },
		function(i, j) { return i % 2 == 0; },
		function(i, j) { return j % 3 == 0; },
		function(i, j) { return (i + j) % 3 == 0; },
		function(i, j) { return (Math.floor(i / 2) + Math.floor(j / 3)) % 2 == 0; },
		function(i, j) { return (i * j) % 2 + (i * j) % 3 == 0; },
		function(i, j) { return ((i * j) % 2 + (i * j) % 3) % 2 == 0; },
		function(i, j) { return ((i * j) % 3 + (i + j) % 2) % 2 == 0; },
		];

	function createBestMatrix(dataList, errorCorrectLevel) {
		var data = errorCorrectLevel.createData(dataList);
		if (! data) return;

		var bestLostPoint = 0;
		var bestPattern = 0;
		for (var i = 0; i < maskFunctions.length; i++) {
			var matrix = new Matrix(data, errorCorrectLevel, true, i);
			var lostPoint = matrix.calculateLostPoint();
			if (i > 0 && lostPoint >= bestLostPoint) continue;
			bestLostPoint = lostPoint;
			bestPattern = i;
		}

		return new Matrix(data, errorCorrectLevel, false, bestPattern);
	}

	function Matrix(data, errorCorrectLevel, test, maskPattern) {
		var sideLength = data.typeNumber * 4 + 17;
		var dots = new Array(sideLength);
		for (var row = 0; row < sideLength; row++) {
			dots[row] = new Array(sideLength);
			for (var col = 0; col < sideLength; col++) {
				dots[row][col] = null;
			}
		}

		setupPositionProbePattern(0, 0);
		setupPositionProbePattern(sideLength - 7, 0);
		setupPositionProbePattern(0, sideLength - 7);
		setupPositionAdjustPatterns();
		setupTimingPattern();
		setupTypeInfo();

		if (data.typeNumber >= 7) setupTypeNumber();
		mapData();

		function setupPositionProbePattern(row, col) {
			for (var r = -1; r <= 7; r++) {
				if (row + r <= -1 || sideLength <= row + r) continue;
				for (var c = -1; c <= 7; c++) {
					if (col + c <= -1 || sideLength <= col + c) continue;
					dots[row + r][col + c] = (0 <= r && r <= 6 && (c == 0 || c == 6)) || (0 <= c && c <= 6 && (r == 0 || r == 6)) || (2 <= r && r <= 4 && 2 <= c && c <= 4);
				}
			}
		}

		function setupTimingPattern() {
			for (var r = 8; r < sideLength - 8; r++) {
				if (dots[r][6] != null) continue;
				dots[r][6] = (r % 2 == 0);
			}

			for (var c = 8; c < sideLength - 8; c++) {
				if (dots[6][c] != null) continue;
				dots[6][c] = (c % 2 == 0);
			}
		}

		function setupPositionAdjustPatterns() {
			var pos = getPatternPosition(data.typeNumber);
			for (var i = 0; i < pos.length; i++) {
				for (var j = 0; j < pos.length; j++) {
					var row = pos[i];
					var col = pos[j];
					if (dots[row][col] != null) continue;
					setupPositionAdjustPattern(row, col);
				}
			}
		}

		function setupPositionAdjustPattern(row, col) {
			for (var r = -2; r <= 2; r++)
				for (var c = -2; c <= 2; c++)
					dots[row + r][col + c] = r == -2 || r == 2 || c == -2 || c == 2 || (r == 0 && c == 0);
		}

		function setupTypeNumber() {
			var bits = getBCHTypeNumber(data.typeNumber);
			for (var i = 0; i < 18; i++) {
				var mod = (! test && ((bits >> i) & 1) == 1);
				dots[Math.floor(i / 3)][i % 3 + sideLength - 8 - 3] = mod;
			}

			for (var i = 0; i < 18; i++) {
				var mod = (! test && ((bits >> i) & 1) == 1);
				dots[i % 3 + sideLength - 8 - 3][Math.floor(i / 3)] = mod;
			}
		}

		function setupTypeInfo() {
			var data = (errorCorrectLevel.code << 3) | maskPattern;
			var bits = getBCHTypeInfo(data);
			for (var i = 0; i < 15; i++) {
				var mod = ! test && ((bits >> i) & 1) == 1;
				if (i < 6) {
					dots[i][8] = mod;
				} else if (i < 8) {
					dots[i + 1][8] = mod;
				} else {
					dots[sideLength - 15 + i][8] = mod;
				}
			}

			for (var i = 0; i < 15; i++) {
				var mod = ! test && ((bits >> i) & 1) == 1;
				if (i < 8) {
					dots[8][sideLength - i - 1] = mod;
				} else if (i < 9) {
					dots[8][15 - i - 1 + 1] = mod;
				} else {
					dots[8][15 - i - 1] = mod;
				}
			}

			dots[sideLength - 8][8] = ! test;
		}

		function mapData() {
			var inc = -1;
			var row = sideLength - 1;
			var bitIndex = 7;
			var byteIndex = 0;
			var maskFunction = maskFunctions[maskPattern];
			for (var col = sideLength - 1; col > 0; col -= 2) {
				if (col == 6) col--;
				while (true) {
					for (var c = 0; c < 2; c++) {
						if (dots[row][col - c] != null) continue;

						var black = false;
						if (byteIndex < data.data.length)
							black = ((data.data[byteIndex] >>> bitIndex) & 1) == 1;

						var mask = maskFunction(row, col - c);
						if (mask) black = ! black;

						dots[row][col - c] = black;
						bitIndex--;
						if (bitIndex == -1) {
							byteIndex++;
							bitIndex = 7;
						}
					}

					row += inc;
					if (row < 0 || sideLength <= row) {
						row -= inc;
						inc = -inc;
						break;
					}
				}
			}
		}

		this.isBlack = function(row, col) {
			return dots[row][col];
		};

		this.getSideLength = function() {
			return sideLength;
		};

		this.calculateLostPoint = function() {
			var lostPoint = 0;
			for (var row = 0; row < sideLength; row++) {
				for (var col = 0; col < sideLength; col++) {
					var sameCount = 0;
					var black = dots[row][col];
					for (var r = -1; r <= 1; r++) {
						if (row + r < 0 || sideLength <= row + r) continue;
						for (var c = -1; c <= 1; c++) {
							if (col + c < 0 || sideLength <= col + c) continue;
							if (r == 0 && c == 0) continue;
							if (black == dots[row + r][col + c]) sameCount++;
						}
					}

					if (sameCount > 5)
						lostPoint += (3 + sameCount - 5);
				}
			}

			for (var row = 0; row < sideLength - 1; row++) {
				for (var col = 0; col < sideLength - 1; col++) {
					var count = 0;
					if (dots[row][col]) count++;
					if (dots[row + 1][col]) count++;
					if (dots[row][col + 1]) count++;
					if (dots[row + 1][col + 1]) count++;

					if (count == 0 || count == 4)
						lostPoint += 3;
				}
			}

			for (var row = 0; row < sideLength; row++)
				for (var col = 0; col < sideLength - 6; col++)
					if (dots[row][col] && ! dots[row][col + 1] && dots[row][col + 2] && dots[row][col + 3] && dots[row][col + 4] && ! dots[row][col + 5] && dots[row][col + 6])
						lostPoint += 40;

			for (var col = 0; col < sideLength; col++)
				for (var row = 0; row < sideLength - 6; row++)
					if (dots[row][col] && ! dots[row + 1][col] && dots[row + 2][col] && dots[row + 3][col] && dots[row + 4][col] && ! dots[row + 5][col] && dots[row + 6][col])
						lostPoint += 40;

			var blackCount = 0;
			for (var col = 0; col < sideLength; col++)
				for (var row = 0; row < sideLength; row++)
					if (dots[row][col])
						blackCount++;

			var ratio = Math.abs(100 * blackCount / sideLength / sideLength - 50) / 5;
			lostPoint += ratio * 10;
			return lostPoint;
		};

		this.toString = function(whiteMarker, blackMarker) {
			if (! whiteMarker) whiteMarker = '  ';
			if (! blackMarker) blackMarker = 'X ';

			var text = [];
			for (var row = 0; row < sideLength; row++) {
				for (var column = 0; column < sideLength; column++)
					text.push(qrCode.isBlack(row, column) ? blackMarker : whiteMarker);

				text.push('\n');
			}

			return text.join('');
		};

		var svgNS = 'http://www.w3.org/2000/svg';

		this.addToSvg = function(parent, whiteColor, blackColor) {
			var sideLength = this.getSideLength();

			var white = document.createElementNS(svgNS, 'rect');
			white.setAttribute('x', 0);
			white.setAttribute('y', 0);
			white.setAttribute('width', sideLength);
			white.setAttribute('height', sideLength);
			white.style.fill = whiteColor;
			parent.appendChild(white);

			var d = '';
			for (var row = 0; row < sideLength; row++)
				for (var column = 0; column < sideLength; column++)
					if (this.isBlack(row, column))
						d += 'M ' + column + ',' + row + ' l 1,0 0,1 -1,0 ';

			var black = document.createElementNS(svgNS, 'path');
			black.setAttribute('d', d);
			black.style.fill = blackColor;
			parent.appendChild(black);
		};

		this.createSvg = function(scaling, whiteColor, blackColor) {
			var sideLength = this.getSideLength();

			var svg = document.createElementNS(svgNS, 'svg');
			svg.setAttribute('width', sideLength * scaling);
			svg.setAttribute('height', sideLength * scaling);
			svg.setAttribute('viewBox', '0 0 ' + sideLength + ' ' + sideLength);
			this.addToSvg(svg, whiteColor, blackColor);
			return svg;
		};

		this.drawToContext2D = function(context, whiteColor, blackColor) {
			var sideLength = this.getSideLength();

			context.fillStyle = whiteColor;
			context.fillRect(0, 0, sideLength, sideLength);

			for (var row = 0; row < sideLength; row++)
				for (var column = 0; column < sideLength; column++)
					if (this.isBlack(row, column))
						context.rect(column, row, 1, 1);

			context.fillStyle = blackColor;
			context.fill();
		};

		this.createPng = function(scale, whiteColor, blackColor, onDone) {
			var sideLength = this.getSideLength();

			var canvas = document.createElement('canvas');
			canvas.width = sideLength * scale;
			canvas.height = sideLength * scale;

			var context = canvas.getContext('2d');
			context.scale(scale, scale);
			this.drawToContext2D(context, whiteColor, blackColor);

			canvas.toBlob(onDone, 'image/png');
		};
	}

	var G15 = 0b10100110111;
	var G18 = 0b1111100100101;
	var G15_MASK = 0b101010000010010;

	var PATTERN_POSITION_TABLE = [
		[],
		[6, 18],
		[6, 22],
		[6, 26],
		[6, 30],
		[6, 34],
		[6, 22, 38],
		[6, 24, 42],
		[6, 26, 46],
		[6, 28, 50],
		[6, 30, 54],
		[6, 32, 58],
		[6, 34, 62],
		[6, 26, 46, 66],
		[6, 26, 48, 70],
		[6, 26, 50, 74],
		[6, 30, 54, 78],
		[6, 30, 56, 82],
		[6, 30, 58, 86],
		[6, 34, 62, 90],
		[6, 28, 50, 72, 94],
		[6, 26, 50, 74, 98],
		[6, 30, 54, 78, 102],
		[6, 28, 54, 80, 106],
		[6, 32, 58, 84, 110],
		[6, 30, 58, 86, 114],
		[6, 34, 62, 90, 118],
		[6, 26, 50, 74, 98, 122],
		[6, 30, 54, 78, 102, 126],
		[6, 26, 52, 78, 104, 130],
		[6, 30, 56, 82, 108, 134],
		[6, 34, 60, 86, 112, 138],
		[6, 30, 58, 86, 114, 142],
		[6, 34, 62, 90, 118, 146],
		[6, 30, 54, 78, 102, 126, 150],
		[6, 24, 50, 76, 102, 128, 154],
		[6, 28, 54, 80, 106, 132, 158],
		[6, 32, 58, 84, 110, 136, 162],
		[6, 26, 54, 82, 110, 138, 166],
		[6, 30, 58, 86, 114, 142, 170]
		];

	function getBCHTypeInfo(data) {
		var d = data << 10;
		while (getBCHDigit(d) - getBCHDigit(G15) >= 0)
			d ^= (G15 << (getBCHDigit(d) - getBCHDigit(G15)));
		return ((data << 10) | d) ^ G15_MASK;
	}

	function getBCHTypeNumber(data) {
		var d = data << 12;
		while (getBCHDigit(d) - getBCHDigit(G18) >= 0)
			d ^= (G18 << (getBCHDigit(d) - getBCHDigit(G18)));
		return (data << 12) | d;
	}

	function getBCHDigit(data) {
		var digit = 0;
		while (data != 0) {
			digit++;
			data >>>= 1;
		}
		return digit;
	}

	function getPatternPosition(typeNumber) {
		return PATTERN_POSITION_TABLE[typeNumber - 1];
	}

	function getErrorCorrectionPolynomial(errorCorrectionLength) {
		var a = new QRPolynomial([1], 0);
		for (var i = 0; i < errorCorrectionLength; i++)
			a = a.multiply(new QRPolynomial([1, gexp(i)], 0));
		return a;
	}

	function QRPolynomial(num, shift) {
		var offset = 0;
		while (offset < num.length && num[offset] == 0) offset++;

		this.num = new Array(num.length - offset + shift);
		for (var i = 0; i < num.length - offset; i++)
			this.num[i] = num[i + offset];

		this.get = function(index) {
			return this.num[index];
		};

		this.getLength = function() {
			return this.num.length;
		};

		this.multiply = function(e) {
			var num = new Array(this.getLength() + e.getLength() - 1);
			for (var i = 0; i < this.getLength(); i++)
				for (var j = 0; j < e.getLength(); j++)
					num[i + j] ^= gexp(glog(this.get(i)) + glog(e.get(j)));

			return new QRPolynomial(num, 0);
		};

		this.mod = function(e) {
			if (this.getLength() - e.getLength() < 0)
				return this;

			var ratio = glog(this.get(0)) - glog(e.get(0));
			var num = new Array(this.getLength());
			for (var i = 0; i < this.getLength(); i++)
				num[i] = this.get(i);

			for (var i = 0; i < e.getLength(); i++)
				num[i] ^= gexp(glog(e.get(i)) + ratio);

			return new QRPolynomial(num, 0).mod(e);
		};
	}

	function glog(n) {
		if (n < 1) throw new Error('glog(' + n + ')');
		return LOG_TABLE[n];
	}

	function gexp(n) {
		while (n < 0) n += 255;
		while (n >= 256) n -= 255;
		return EXP_TABLE[n];
	}

	var EXP_TABLE = new Array(256);
	var LOG_TABLE = new Array(256);

	for (var i = 0; i < 8; i++) {
		EXP_TABLE[i] = 1 << i;
	}

	for (var i = 8; i < 256; i++) {
		EXP_TABLE[i] = EXP_TABLE[i - 4] ^ EXP_TABLE[i - 5] ^ EXP_TABLE[i - 6] ^ EXP_TABLE[i - 8];
	}

	for (var i = 0; i < 255; i++) {
		LOG_TABLE[EXP_TABLE[i]] = i;
	}

	this.forText = function(text, redundancy) {
		if (! redundancy) redundancy = MEDIUM_REDUNDANCY;
		var data = new TextData(text);
		return createBestMatrix([data], redundancy);
	};
}
